@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2017-2020
+Copyright (C) 2017-2021
 Martijn Braam, Clayton Craft <clayton@craftyguy.net>, et al.
 
 This file is part of osk-sdl.
@@ -21,16 +21,31 @@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #include "keyboard.h"
 #include "draw_helpers.h"
 
-Keyboard::Keyboard(int pos, int targetPos, int width, int height, Config *config)
+Keyboard::Keyboard(int pos, int targetPos, int width, int height, Config *config, SDL_Haptic *haptic)
 	: position(static_cast<float>(pos))
 	, targetPosition(static_cast<float>(targetPos))
 	, keyboardWidth(width)
 	, keyboardHeight(height)
 	, config(config)
+	, haptic(haptic)
 {
 	lastAnimTicks = SDL_GetTicks();
 }
 
+void Keyboard::cleanup()
+{
+	for (auto &layer : keyboard) {
+		if (layer.texture) {
+			SDL_DestroyTexture(layer.texture);
+			layer.texture = nullptr;
+		}
+		if (layer.highlightedTexture) {
+			SDL_DestroyTexture(layer.highlightedTexture);
+			layer.highlightedTexture = nullptr;
+		}
+	}
+}
+
 int Keyboard::init(SDL_Renderer *renderer)
 {
 	loadKeymap();
@@ -70,6 +85,12 @@ void Keyboard::setTargetPosition(float p)
 
 void Keyboard::updateAnimations()
 {
+	// If animations are disabled, just jump straight to target
+	if (!config->animations) {
+		position = targetPosition;
+		return;
+	}
+
 	const int animStep = 20; // 20ms -> 50 FPS
 	const int maxFallBehindSteps = 20;
 	int now = SDL_GetTicks();
@@ -83,20 +104,15 @@ void Keyboard::updateAnimations()
 	while (lastAnimTicks < now) {
 		// Vertical keyboard movement:
 		if (fabs(position - targetPosition) > 0.01) {
-			if (!(config->animations)) {
-				// If animations are disabled, just jump to target:
-				position = targetPosition;
-			} else {
-				// Gradually update the position:
-				if (position > targetPosition) {
-					position -= fmax(0.1, position - targetPosition) / 8;
-					if (position < targetPosition)
-						position = targetPosition;
-				} else if (position < targetPosition) {
-					position += fmax(0.1, targetPosition - position) / 8;
-					if (position > targetPosition)
-						position = targetPosition;
-				}
+			// Gradually update the position:
+			if (position > targetPosition) {
+				position -= fmax(0.1, position - targetPosition) / 8;
+				if (position < targetPosition)
+					position = targetPosition;
+			} else if (position < targetPosition) {
+				position += fmax(0.1, targetPosition - position) / 8;
+				if (position > targetPosition)
+					position = targetPosition;
 			}
 		} else {
 			position = targetPosition;
@@ -216,6 +232,7 @@ void Keyboard::drawRow(SDL_Surface *surface, std::vector<touchArea> &keyVector,
 		keyCapRect.w = keyRect.w;
 		keyCapRect.h = keyRect.h;
 		SDL_BlitSurface(textSurface, nullptr, surface, &keyCapRect);
+		SDL_FreeSurface(textSurface);
 
 		i++;
 	}
@@ -261,12 +278,13 @@ void Keyboard::drawKey(SDL_Surface *surface, std::vector<touchArea> &keyVector,
 	keyCapRect.w = keyRect.w;
 	keyCapRect.h = keyRect.h;
 	SDL_BlitSurface(textSurface, nullptr, surface, &keyCapRect);
+	SDL_FreeSurface(textSurface);
 }
 
 SDL_Surface *Keyboard::makeKeyboard(KeyboardLayer *layer, bool isHighlighted) const
 {
 	SDL_Surface *surface;
-	Uint32 rmask, gmask, bmask, amask;
+	Uint32 rmask, gmask, bmask;
 
 	/* SDL interprets each pixel as a 32-bit number, so our masks must depend
 	   on the endianness (byte order) of the machine */
@@ -274,16 +292,13 @@ SDL_Surface *Keyboard::makeKeyboard(KeyboardLayer *layer, bool isHighlighted) co
 	rmask = 0xff000000;
 	gmask = 0x00ff0000;
 	bmask = 0x0000ff00;
-	amask = 0x000000ff;
 #else
 	rmask = 0x000000ff;
 	gmask = 0x0000ff00;
 	bmask = 0x00ff0000;
-	amask = 0xff000000;
 #endif
 
-	surface = SDL_CreateRGBSurface(SDL_SWSURFACE, keyboardWidth, keyboardHeight, 32, rmask, gmask,
-		bmask, amask);
+	surface = SDL_CreateRGBSurface(SDL_SWSURFACE, keyboardWidth, keyboardHeight, 32, rmask, gmask, bmask, 0);
 
 	if (surface == nullptr) {
 		SDL_LogError(SDL_LOG_CATEGORY_VIDEO, "CreateRGBSurface failed: %s", SDL_GetError());
@@ -307,11 +322,6 @@ SDL_Surface *Keyboard::makeKeyboard(KeyboardLayer *layer, bool isHighlighted) co
 		rowHeight = keyboardHeight / rowCount;
 	}
 
-	if (TTF_Init() == -1) {
-		SDL_LogError(SDL_LOG_CATEGORY_VIDEO, "TTF_Init: %s", TTF_GetError());
-		return nullptr;
-	}
-
 	TTF_Font *font = TTF_OpenFont(config->keyboardFont.c_str(), config->keyboardFontSize);
 	if (!font) {
 		SDL_LogError(SDL_LOG_CATEGORY_VIDEO, "TTF_OpenFont: %s", TTF_GetError());
@@ -338,7 +348,7 @@ SDL_Surface *Keyboard::makeKeyboard(KeyboardLayer *layer, bool isHighlighted) co
 			x = keyboardWidth / 20 + colw * 2;
 		argb keyBackground = i == 0 ? keyBackgroundOther : keyBackgroundLetter;
 		drawRow(surface, layer->keyVector, x, y, rowKeyWidth,
-			rowHeight, layer->rows[i], keyboardWidth / 100, font, isHighlighted, true, keyForeground, keyBackground);
+			rowHeight, layer->rows[i], keyboardWidth / 100, font, isHighlighted, config->keyPreview, keyForeground, keyBackground);
 		y += rowHeight;
 		i++;
 	}
@@ -387,12 +397,14 @@ SDL_Surface *Keyboard::makeKeyboard(KeyboardLayer *layer, bool isHighlighted) co
 
 	char period[] = ".";
 	drawKey(surface, layer->keyVector, colw * 13, y, colw * 2, rowHeight,
-		period, KEYCAP_PERIOD, keyboardWidth / 100, font, isHighlighted, true, keyForeground, keyBackgroundOther);
+		period, KEYCAP_PERIOD, keyboardWidth / 100, font, isHighlighted, config->keyPreview, keyForeground, keyBackgroundOther);
 
 	char enter[] = "OK";
 	drawKey(surface, layer->keyVector, colw * 15, y, colw * 5, rowHeight,
 		enter, KEYCAP_RETURN, keyboardWidth / 100, font, isHighlighted, false, keyForeground, keyBackgroundReturn);
 
+	TTF_CloseFont(font);
+
 	return surface;
 }
 
@@ -497,3 +509,10 @@ touchArea Keyboard::getHighlightedKey()
 {
 	return highlightedKey;
 }
+
+void Keyboard::hapticRumble()
+{
+	if (haptic && config->keyVibrateDuration) {
+		SDL_HapticRumblePlay(haptic, 1, config->keyVibrateDuration);
+	}
+}
